#! /usr/bin/python

import sys

## CODE = INSTRUCTIONS.
## IF = MEMORY|ISINDEX|COMPARE '?' INSTRUCTIONS ':' INSTRUCTIONS.
## COMPARE = ('+'|'-'|'*'|'/') MEMORY '=' (DECINT|MEMORY).
## ISINDEX = '~' (DECINT|MEMORY).
## INSTRUCTIONS = INSTRUCTION { ',' INSTRUCTION }.
## INSTRUCTION = RECURSION|MEMSET|INPUT|OUTPUT|IF.
## OUTPUT = ('>'|'!') MEMORY.
## INPUT = '<' MEMORY.
## RECURSION = '=' DECINT.
## MEMSET = MEMORY ('='|'+'|'*'|'/'|'-') DECINT|STRING|MEMORY.
## STRING = '"' ascii '"'.
## MEMORY = '.' DECINT ['!' (DECINT|MEMORY)]
## DECINT = '_' | '$' | DECDIGIT {DECDIGIT}.
## DECDIGIT = '0' .. '9'.

def DebugFunc(*args):
    print " ".join(map(str,args))

def Quiet(*args):
    pass

Debug = Quiet

class Memory:
    def __init__(self):
        self.memory = {}

    def __getitem__(self,index):
        offset = index[0]
        if type(offset) == type([]):
            offset = self.memory[offset[0]]
        if not self.memory.has_key(offset):
            self.memory[offset] = 0
        result = self.memory[offset]
        if len(index)==2:
            offset = index[1]
            if type(offset) == type(0):
                result = result[offset]
            else:
                result = result[memory[offset]]
        Debug("GET memory",offset,"=",result)
        return result

    def __setitem__(self,index,value):
        offset = index[0]
        if type(offset) == type([]):
            offset = self.memory[offset[0]]
        self.memory[offset] = value
        Debug("SET memory[",offset,"]=",value)
        return 0
    
memory = Memory()

class Instruction:

    def Value(self,data,index):    
        if type(data) == type(0):
            return data
        elif type(data) == type(""):
            return map(ord,list(data)) + [0]
        elif type(data) == type([]):
            return memory[data]
        elif type(data) == type(self):
            return data.Execute(index)
        else:
            raise "ERROR, expected type int, string or Instruction for assignment but got %s instead" % str(type(self.source))
        
    def Assign(self,index):
        memory[self.target] = self.Value(self.source,index)
        return memory[self.target]

    def Add(self,index):
        value = self.Value(self.source,index)
        Debug("Add",value,"to",memory[self.target])
        memory[self.target] += value
        return memory[self.target]

    def Subtract(self,index):
        memory[self.target] -= self.Value(self.source,index)
        return memory[self.target]

    def Multiply(self,index):
        memory[self.target] *= self.Value(self.source,index)
        return memory[self.target]

    def Divide(self,index):
        memory[self.target] /= self.Value(self.source,index)
        return memory[self.target]

    def LessThan(self,index):
        a = self.Value(self.a,index)
        b = self.Value(self.b,index)
        Debug("Test if",a,"is less than",b)
        return a<b

    def Equal(self,index):
        return self.Value(self.a,index) == self.Value(self.b,index)

    def GreaterThan(self,index):
        return self.Value(self.a,index) > self.Value(self.b,index)

    def NotEqual(self,index):
        return not (self.Value(self.a,index) == self.Value(self.b,index))
    
    def Recurse(self,index):
        global root_sequence
        Debug("Call self recursively with index",self.index)
        return ExecuteInstructionSequence(root_sequence,self.index)

    def Output(self,index):
        return 1, sys.stdout.write(chr(memory[self.address]))

    def OutputInteger(self,index):
        print memory[self.address]
        
    def Constant(self,index):
        return self.value

    def Input(self,index):
        print "Input memoy at", self.address
        raise "ERROR, not implemented yet"

    def If(self,index):
        if type(self.address) == type(self):
            result = self.address.Execute(index)
        else:
            result = memory[self.address]
        Debug("IF-Expression is",result)
        if result:
            return ExecuteInstructionSequence(self.if_true,index)
        else:
            return ExecuteInstructionSequence(self.if_false,index)

    def IsIndex(self,index):
        return index == self.Value(self.value,index)

def ExecuteInstructionSequence(instructions,index):
    result = 0
    for instruction in instructions:
        result = instruction.Execute(index)
    return result

class JustifParser:
    def Parse(self,expression):
        self.expression = expression
        self.pos = 0
        self.nums = [-1,-1]
        return self.INSTRUCTIONS()

    def SkipWhitespaces(self):
        try:
            while self.expression[self.pos] in " \r\nABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz":
                self.pos += 1
        except:
            pass

    def PushPos(self):
        return self.pos, self.nums[:]

    def PopPos(self,s):
        self.pos = s[0]
        self.nums = s[1]        

    def GetChar(self,pos=0):
        self.SkipWhitespaces()
        try:
            return self.expression[self.pos+pos]
        except:
            return 0    

    def SkipChar(self,expression):
        self.SkipWhitespaces()
        try:
            if self.expression[self.pos:self.pos+len(expression)] == expression:
                self.pos += len(expression)
                return 1
        except:
            pass

    def ISINDEX(self):
        pos = self.PushPos()
        if self.SkipChar('~'):
            value = self.DECINT()
            if value is None:
                value = self.MEMORY()
            if value is not None:
                result = Instruction()
                result.value = value
                result.Execute = result.IsIndex
                return result                
        self.PopPos(pos)

    def COMPARE(self):
        pos = self.PushPos()
        c = self.GetChar()
        if c in '+-*/':
            self.pos += 1
            m = self.MEMORY()
            if m is not None and self.SkipChar('='):
                v = self.DECINT()
                if v is None:
                    v = self.MEMORY()
                if v is not None:
                    result = Instruction()
                    result.a = m
                    result.b = v
                    result.Execute = [result.LessThan,result.Equal,result.GreaterThan,result.NotEqual]['+-*/'.index(c)]
                    return result                     
        self.PopPos(pos)

    def IF(self):        
        pos = self.PushPos()
        m = self.MEMORY()
        if m is None:
            m = self.COMPARE()
        if m is None:
            m = self.RECURSION()
        if m is None:
            m = self.ISINDEX()
        if m is not None and self.SkipChar('?'):
            if_true = self.INSTRUCTIONS()
            if if_true and self.SkipChar(':'):
                if_false = self.INSTRUCTIONS()
                if if_false:
                    result = Instruction()
                    result.address = m
                    result.if_true = if_true
                    result.if_false = if_false
                    result.Execute = result.If
                    return result
        self.PopPos(pos)

    def INPUT(self):
        pos = self.PushPos()
        if self.SkipChar('<'):
            address = self.MEMORY()
            if address:
                result = Instruction()
                result.address = address
                result.Execute = result.Input
                return result
        self.PopPos(pos)
        
    def OUTPUT(self):
        pos = self.PushPos()
        if self.SkipChar('>'):
            address = self.MEMORY()
            if address:
                result = Instruction()
                result.address = address
                result.Execute = result.Output
                return result
        elif self.SkipChar('!'):
            address = self.MEMORY()
            if address:
                result = Instruction()
                result.address = address
                result.Execute = result.OutputInteger
                return result            
        self.PopPos(pos)

    def INSTRUCTIONS(self):
        pos = self.PushPos()
        instruction = self.INSTRUCTION()
        if instruction:
            result = []
            while 1:
                result.append(instruction)
                if self.SkipChar(','):
                    instruction = self.INSTRUCTION()
                    if instruction:
                        continue
                break
            return result
        self.PopPos(pos)

    def INSTRUCTION(self):
        pos = self.PushPos()
        for function in (self.IF,self.COMPARE,self.RECURSION,self.MEMSET,self.OUTPUT,self.INPUT,self.CONSTANT):
            instruction = function()
            if instruction is not None:
                return instruction
        self.PopPos(pos)

    def CONSTANT(self):
        decint = self.DECINT()
        if decint is not None:
            result = Instruction()
            result.value = decint
            result.Execute = result.Constant
            return result

    def STRING(self):
        if self.SkipChar('"'):
            startpos = self.pos
            while 1:
                c = self.GetChar()
                if c == 0:
                    raise "ERROR, expected end-of-string"
                self.pos += 1
                if c == '"':
                    return self.expression[startpos:self.pos-1]
        
    def RECURSION(self):
        pos = self.PushPos()
        if self.SkipChar('='):
            index = self.DECINT()
            if index is not None:
                result = Instruction()
                result.index = index
                result.Execute = result.Recurse
                return result            
        self.PopPos(pos)

    def MEMSET(self):
        pos = self.PushPos()
        m = self.MEMORY()
        if m:
            c = self.GetChar()
            if c in '=+-*/':
                self.pos += 1
                d = self.DECINT()
                if d is None:
                    d = self.STRING()
                if d is None:
                    d = self.MEMORY()
                if d is not None:
                    result = Instruction()
                    result.source = d
                    result.operation = c
                    result.target = m
                    result.Execute = [result.Assign,result.Add,result.Subtract,result.Multiply,result.Divide]['=+-*/'.index(c)]
                    return result 
        self.PopPos(pos)

    def MEMORY(self):
        pos = self.PushPos()
        if self.SkipChar('.'):
            d = self.DECINT()
            if d is None:
                d = self.MEMORY()
            if d is not None:
                result = [d]
                while self.SkipChar('!'):
                    number = self.DECINT()
                    if number is None:
                        number = self.MEMORY()
                        if number is None:
                            raise "Syntax Error: expected decint after '!'"
                    result.append(number)
                return result
        self.PopPos(pos)          

    def DECINT(self):
        number = 0
        if self.SkipChar('_'):
            return self.nums[0]
        if self.SkipChar('$'):
            return self.nums[1]
        success, digit = self.DECDIGIT()
        if success:
            while success:
                number *= 10
                number += digit
                success, digit = self.DECDIGIT()
            self.nums[1] = self.nums[0]
            self.nums[0] = number
            return number

    def DECDIGIT(self):
        c = self.GetChar()
        if c >= '0' and c <= '9':
            self.pos += 1
            return 1, ord(c)-ord('0')
        return 0, 0

if __name__=='__main__':
    if len(sys.argv) == 2:
        j = JustifParser()
        root_sequence = j.Parse(open(sys.argv[1]).read())
        if root_sequence is not None:
            ExecuteInstructionSequence(root_sequence,1)
        else:
            print "ERROR, unable to parse", sys.argv[1]
    else:
        print "USAGE: justif.py <filename>"
